#include <Arduino.h>
#include <string>
#include <Wire.h>
#include <Adafruit_NeoPixel.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <hd44780.h>
#include <hd44780ioClass/hd44780_I2Cexp.h>
#include <Adafruit_Sensor.h>
#include <DHT.h>
#include <ESP32Servo.h>
#include <WebServer.h>
#include <ArduinoJson.h>


// set up your wifi name
#define WIFI_SSID "ACEBOTT"

// set up your wifi password
#define WIFI_PASSWORD "12345678"

//  set up WEB server
#define WEB_SERVER_PORT 80
#define ALL_SENSOR_HTTP_PATH "/all"
#define WIFI_CONNECT_MAX_TIMEOUT 5 // 连接超时秒数
WebServer server(WEB_SERVER_PORT); // WIFI服务器
int wifiConnectRetryNum = 0;       // WIFI连接重试计数
int wifiConnectState = false;      // WIFI连接状态

// servo of door
#define DOOR_SERVO_PIN 4
#define DOOR_SERVO_ANGLE_OPEN 100
#define DOOR_SERVO_ANGLE_CLOSE 5
#define DOOR_TEXT_ON "door_open"
#define DOOR_TEXT_OFF "door_close"
#define DOOR_HTTP_PATH "/door"

// servo of window
#define WINDOW_SERVO_PIN 13
#define WINDOW_SERVO_ANGLE_OPEN 123
#define WINDOW_SERVO_ANGLE_CLOSE 60
#define WINDOW_TEXT_ON "window_open"
#define WINDOW_TEXT_OFF "window_close"
#define WINDOW_HTTP_PATH "/window"

// laser
#define LASER_PIN 23
#define LASER_TEXT_ON "laser_on"
#define LASER_TEXT_OFF "laser_off"
#define LASER_HTTP_PATH "/laser"

// PIR sensor
#define HUMAN_PIN 18

// LED
#define LED_PIN 19
#define LED_TEXT_ON "LED_on"
#define LED_TEXT_OFF "LED_off"
#define LED_HTTP_PATH "/led"

// light sensor
#define BRIGHTNESS_PIN 32
#define BRIGHTNESS_MAX_VALUE 3000

// button
#define BUTTON_PIN 26

// DHT11
#define DHT11_PIN 25

// raindrop sensor
#define RAINDROP_PIN 39
#define RAINDROP_MAX_VALUE 2000

// LCD 1602
#define DISPLAY_HD44780_I2C_ADDRESS 0x27
#define DISPLAY_HD44780_ROWS 2
#define DISPLAY_HD44780_COLS 16

// WS2812 RGB strip
#define WS2812_PIN 33
#define WS2812_NUM 9
#define WS2812_TEXT_ON "RGB_on"
#define WS2812_TEXT_OFF "RGB_off"
#define WS2812_HTTP_PATH "/rgb"
#define WS2812_COLOR_BLUE pixels.Color(0, 0, 255)
#define WS2812_COLOR_BLACK pixels.Color(0, 0, 0)

// creat a DHT11 object
DHT dht(DHT11_PIN, DHT11);

// creat a ws2812 object
Adafruit_NeoPixel pixels(WS2812_NUM, WS2812_PIN, NEO_GRB + NEO_KHZ800);

// creat a lcd object
hd44780_I2Cexp lcd(DISPLAY_HD44780_I2C_ADDRESS, DISPLAY_HD44780_ROWS, DISPLAY_HD44780_COLS);

// creat a servo object of window
Servo windowServo;

// creat a servo object of door
Servo doorServo;

String historyVal = "";
int historyCol = -1;
int historyRow = -1;
// display texts on LCD1602
void drawStr(uint8_t col, uint8_t row, String str)
{
  if (historyVal != str || historyCol != col || historyRow != row)
  {
    lcd.setCursor(col, row);
    lcd.print("               ");
    lcd.setCursor(col, row);
    lcd.print(str);
    historyVal = str;
    historyCol = col;
    historyRow = row;
  }
}

// get PIR sensor value
int getHumanValue()
{
  return digitalRead(HUMAN_PIN);
}


// control LED
void setLedValue(int val)
{
  digitalWrite(LED_PIN, val);
}


// control laser
void setLaserValue(int val)
{
  digitalWrite(LASER_PIN, val);
}

// get temperature value
float getTemperatureValue()
{
  return dht.readTemperature();
}

// get humidity value
float getHumidityValue()
{
  return dht.readHumidity();
}

// get light value
float getBrightnessValue()
{
  return analogRead(BRIGHTNESS_PIN);
}

// get raindrop sensor value
float getRaindropValue()
{
  return analogRead(RAINDROP_PIN);
}

// display temperature value and humidity value on LCD
void humitureLoop()
{
  drawStr(0, 0, "T:" + String(dht.readTemperature()) + " H:" + String(dht.readHumidity()));
}

// initial sensors and modules
void initializeSensor()
{
  pixels.begin();
  lcd.begin(DISPLAY_HD44780_COLS, DISPLAY_HD44780_ROWS);
  lcd.backlight();
  lcd.clear();

  dht.begin();
  Wire.begin();


  windowServo.attach(WINDOW_SERVO_PIN);
  windowServo.write(WINDOW_SERVO_ANGLE_OPEN);
  doorServo.attach(DOOR_SERVO_PIN);
  doorServo.write(DOOR_SERVO_ANGLE_CLOSE);

  pinMode(HUMAN_PIN, INPUT);
  pinMode(BRIGHTNESS_PIN, INPUT);
  pinMode(LED_PIN, OUTPUT);
  pinMode(LASER_PIN, OUTPUT);
}

// connect to wifi
void connectWifi()
{
  WiFi.setTxPower(WIFI_POWER_19_5dBm);
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
  while (WiFi.status() != WL_CONNECTED)
  {
    delay(1000);
    Serial.println(".");
    wifiConnectRetryNum++;
    if (wifiConnectRetryNum > WIFI_CONNECT_MAX_TIMEOUT)
    {
      break;
    }
  }
  if (WiFi.status() != WL_CONNECTED)
  {
    wifiConnectState = false;
  }
  else
  {
    wifiConnectState = true;
  }
  Serial.print("IP:");
  Serial.println(WiFi.localIP());
}

// set color of RGB strip
void ws2812ShowColor(int color)
{
  for (int i = 0; i < WS2812_NUM; i++)
  {
    pixels.setPixelColor(i, color);
  }
  pixels.show();
}

// set the R,G,B value of RGB strip
void ws2812ShowRGB(int r, int g, int b)
{
  for (int i = 0; i < WS2812_NUM; i++)
  {
    pixels.setPixelColor(i, pixels.Color(r, g, b));
  }
  pixels.show();
}

void handleRoot()
{
  server.send(200, "text/plain", WiFi.localIP().toString());
}

void handleConfig()
{
  JsonDocument doc;
  doc["appCallInterval"] = 1000;
  String jsonString;
  serializeJson(doc, jsonString);
  server.send(200, "text/plain", jsonString);
}

void handleAll()
{
  JsonDocument doc;
  doc["Photoresistor"] = getBrightnessValue();
  doc["Human"] = getHumanValue();
  doc["Raindrop"] = getRaindropValue();
  doc["Temperature"] = getTemperatureValue();
  doc["Humidity"] = getHumidityValue();
  String jsonString;
  serializeJson(doc, jsonString);
  server.send(200, "text/plain", jsonString);
}

void handleDoor()
{
  String level = server.arg("level");
  Serial.print("level:");
  Serial.println(level);
  if (level.toInt())
  {
    doorServo.write(DOOR_SERVO_ANGLE_OPEN);
  }
  else
  {
    doorServo.write(DOOR_SERVO_ANGLE_CLOSE);
  }
  server.send(200, "text/plain", "ok");
}

void handleWindow()
{
  String level = server.arg("level");
  Serial.print("level:");
  Serial.println(level);
  if (level.toInt())
  {
    windowServo.write(WINDOW_SERVO_ANGLE_OPEN);
  }
  else
  {
    windowServo.write(WINDOW_SERVO_ANGLE_CLOSE);
  }
  server.send(200, "text/plain", "ok");
}

void handleLaser()
{
  String level = server.arg("level");
  Serial.print("level:");
  Serial.println(level);
  setLaserValue(level.toInt());
  server.send(200, "text/plain", "ok");
}

void handleLED()
{
  String level = server.arg("level");
  Serial.print("level:");
  Serial.println(level);
  setLedValue(level.toInt());
  server.send(200, "text/plain", "ok");
}

void handleRGB()
{
  String r = server.arg("r");
  String g = server.arg("g");
  String b = server.arg("b");
  Serial.print("r:");
  Serial.println(r);
  Serial.print("g:");
  Serial.println(g);
  Serial.print("b:");
  Serial.println(b);
  ws2812ShowRGB(r.toInt(), g.toInt(), b.toInt());
  server.send(200, "text/plain", "ok");
}
void setup()
{
  Serial.begin(115200);
  initializeSensor();
  connectWifi();
  if (wifiConnectState)
  {
    drawStr(0, 1, WiFi.localIP().toString());

    server.on("/", handleRoot);
    server.on(ALL_SENSOR_HTTP_PATH, handleAll);
    server.on(DOOR_HTTP_PATH, handleDoor);
    server.on(WINDOW_HTTP_PATH, handleWindow);
    server.on(LASER_HTTP_PATH, handleLaser);
    server.on(LED_HTTP_PATH, handleLED);
    server.on(WS2812_HTTP_PATH, handleRGB);
    server.begin();
  }
  else
  {
    drawStr(0, 1, "Wifi failed!");
  }
}

void loop()
{
  server.handleClient();
}